package net.knarcraft.bookswithoutborders.listener;

import net.knarcraft.bookswithoutborders.config.BooksWithoutBordersConfig;
import net.knarcraft.bookswithoutborders.utility.IntegerToRomanConverter;
import net.md_5.bungee.api.ChatColor;
import org.bukkit.block.ChiseledBookshelf;
import org.bukkit.enchantments.Enchantment;
import org.bukkit.entity.Player;
import org.bukkit.event.Event;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.block.Action;
import org.bukkit.event.player.PlayerInteractEvent;
import org.bukkit.inventory.ChiseledBookshelfInventory;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.BookMeta;
import org.bukkit.inventory.meta.EnchantmentStorageMeta;
import org.bukkit.inventory.meta.ItemMeta;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import static net.knarcraft.bookswithoutborders.config.BooksWithoutBordersConfig.getSuccessColor;

/**
 * A listener for bookshelf clicking
 */
public class BookshelfListener implements Listener {

    @EventHandler
    public void onBookshelfClick(PlayerInteractEvent event) {
        Player player = event.getPlayer();

        // If left-clicking a chiseled bookshelf and sneaking, display contents
        if (!event.hasBlock() || !(event.getClickedBlock().getState() instanceof ChiseledBookshelf chiseledBookshelf) ||
                event.getAction() != Action.LEFT_CLICK_BLOCK || !player.isSneaking()) {
            return;
        }

        // Check if bookshelf peeking is enabled, and the player can peek
        if (!BooksWithoutBordersConfig.getEnableBookshelfPeeking() ||
                !event.getPlayer().hasPermission("bookswithoutborders.peekbookshelf")) {
            return;
        }

        event.setUseInteractedBlock(Event.Result.DENY);
        event.setUseItemInHand(Event.Result.DENY);

        ChiseledBookshelfInventory bookshelfInventory = chiseledBookshelf.getInventory();
        player.sendMessage(getBookshelfDescription(bookshelfInventory));
    }

    /**
     * Gets the description for a bookshelf's contents
     *
     * @param bookshelfInventory <p>The inventory of the bookshelf to describe</p>
     * @return <p>A textual description of the bookshelf's contents</p>
     */
    private String getBookshelfDescription(ChiseledBookshelfInventory bookshelfInventory) {
        StringBuilder builder = new StringBuilder(getSuccessColor() + "Books in shelf:");
        for (ItemStack itemStack : bookshelfInventory.getStorageContents()) {
            if (itemStack == null) {
                continue;
            }
            ItemMeta meta = itemStack.getItemMeta();
            builder.append("\n ").append(ChatColor.GRAY).append(" - ");
            if (meta instanceof BookMeta bookMeta) {
                builder.append(getBookDescription(bookMeta));
            } else if (meta instanceof EnchantmentStorageMeta enchantmentStorageMeta) {
                builder.append(getEnchantedBookDescription(enchantmentStorageMeta));
            } else if (meta != null) {
                builder.append(getPlainBookDescription(meta));
            }
        }
        return builder.toString();
    }

    /**
     * Gets the description of a plain (enchant-able) book
     *
     * @param itemMeta <p>The metadata for the book to describe</p>
     * @return <p>The description of the book</p>
     */
    private String getPlainBookDescription(ItemMeta itemMeta) {
        String name = itemMeta.getDisplayName();
        if (name.isEmpty()) {
            name = "Plain book";
        }
        return name;
    }

    /**
     * Gets the description of a book
     *
     * @param bookMeta <p>The metadata for the book to describe</p>
     * @return <p>The book's description</p>
     */
    private String getBookDescription(BookMeta bookMeta) {
        String title;
        String author;
        if (!bookMeta.hasTitle() || bookMeta.getTitle() == null) {
            title = "Untitled";
        } else {
            title = bookMeta.getTitle();
        }
        if (!bookMeta.hasAuthor() || bookMeta.getAuthor() == null) {
            author = "Unknown";
        } else {
            author = bookMeta.getAuthor();
        }
        return title + " by " + author;
    }

    /**
     * Gets the description of an enchanted book
     *
     * @param enchantmentStorageMeta <p>The metadata for the enchanted book to describe</p>
     * @return <p>The enchanted book's description</p>
     */
    private String getEnchantedBookDescription(EnchantmentStorageMeta enchantmentStorageMeta) {
        StringBuilder builder = new StringBuilder();
        builder.append("Enchanted book (");
        Map<Enchantment, Integer> enchantmentMap = enchantmentStorageMeta.getStoredEnchants();
        List<String> enchantments = new ArrayList<>(enchantmentMap.size());
        for (Map.Entry<Enchantment, Integer> enchantmentEntry : enchantmentMap.entrySet()) {
            enchantments.add(getEnchantmentName(enchantmentEntry.getKey()) + " " +
                    IntegerToRomanConverter.getRomanNumber(enchantmentEntry.getValue()));
        }
        builder.append(String.join(", ", enchantments));
        builder.append(")");
        return builder.toString();
    }

    /**
     * Gets a prettified name of an enchantment
     *
     * @param enchantment <p>The enchantment to get the name of</p>
     * @return <p>The prettified enchantment name</p>
     */
    private String getEnchantmentName(Enchantment enchantment) {
        return uppercaseFirst(enchantment.getKey().getKey().replace("_", " "));
    }

    /**
     * Turns the first character of each of a string's words into uppercase
     *
     * @param input <p>The input to uppercase</p>
     * @return <p>The input string with more uppercase</p>
     */
    private String uppercaseFirst(String input) {
        String[] parts = input.split(" ");
        for (int i = 0; i < parts.length; i++) {
            parts[i] = parts[i].substring(0, 1).toUpperCase() + parts[i].substring(1);
        }
        return String.join(" ", parts);
    }

}
