package net.knarcraft.bookswithoutborders.utility;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import net.knarcraft.bookswithoutborders.config.BooksWithoutBordersConfig;
import net.milkbowl.vault.economy.Economy;
import org.bukkit.Material;
import org.bukkit.Server;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.PlayerInventory;
import org.bukkit.inventory.meta.BookMeta;
import org.bukkit.plugin.Plugin;
import org.bukkit.plugin.RegisteredServiceProvider;
import org.bukkit.plugin.ServicesManager;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 * Helper class for economy related functions
 */
public final class EconomyHelper {

    private static Economy economy;

    private EconomyHelper() {
    }

    /**
     * Gets an economy instance for making transactions
     *
     * @return <p>An economy instance, or null if it's not initialized</p>
     */
    public static Economy getEconomy() {
        return economy;
    }

    /**
     * Tries to set up economy
     *
     * @return <p>True if economy is set up and enabled</p>
     */
    public static boolean setupEconomy() {
        if (economy != null) {
            return true;
        }

        Server server = BooksWithoutBorders.getInstance().getServer();
        Plugin plugin = server.getPluginManager().getPlugin("Vault");
        ServicesManager servicesManager = server.getServicesManager();

        if (plugin != null) {
            RegisteredServiceProvider<Economy> economyProvider = servicesManager.getRegistration(Economy.class);
            if (economyProvider != null) {
                economy = economyProvider.getProvider();
            }
        }

        return (economy != null);
    }

    /**
     * Makes the player pay for printing a given number of books
     *
     * @param player    <p>The player printing the books</p>
     * @param numCopies <p>The number of copies the player is trying to print</p>
     * @return <p>True if the player cannot pay for the printing of the books</p>
     */
    public static boolean cannotPayForBookPrinting(Player player, int numCopies) {
        //BookPriceQuantity: How many items are required to pay for each book
        //BookPriceType: Which item is used to pay for the books. AIR = use economy
        Material bookCurrency = BooksWithoutBordersConfig.getBookPriceType();
        double cost = BooksWithoutBordersConfig.getBookPriceQuantity() * numCopies;
        int itemCost = (int) cost;

        if (bookCurrency == Material.AIR) {
            return !EconomyHelper.payForBookPrintingEconomy(player, cost, numCopies);
        } else {
            if (bookCurrency == Material.WRITABLE_BOOK) {
                //Writable books are treated as a special case to prevent WIP books from being used
                return !takeWritableBookPayment(player, itemCost);
            } else {
                if (player.getInventory().contains(bookCurrency, itemCost)) {
                    payForBookPrintingItem(player, itemCost);
                    return false;
                } else {
                    BooksWithoutBorders.sendErrorMessage(player, itemCost + " " + bookCurrency +
                            "(s) are required for this command!");
                    return true;
                }
            }
        }
    }

    /**
     * Takes a writable book payment, ignoring any books containing text
     *
     * @param player   <p>The player to take payment from</p>
     * @param itemCost <p>The number of writable books to pay</p>
     * @return <p>True if the payment was successful</p>
     */
    private static boolean takeWritableBookPayment(Player player, int itemCost) {
        List<ItemStack> books = getPlayersEmptyBooks(player);
        if (countItems(books) < itemCost) {
            BooksWithoutBorders.sendErrorMessage(player, itemCost + " empty " + Material.WRITABLE_BOOK +
                    "(s) are required for this command!");
            return false;
        } else {
            int clearedAmount = 0;
            for (ItemStack itemStack : books) {
                if (itemStack.getAmount() > itemCost) {
                    //If encountering an item stack with more than enough books, remove the necessary books
                    itemStack.setAmount(itemStack.getAmount() - itemCost);
                    return true;
                } else {
                    clearedAmount += itemStack.getAmount();
                    player.getInventory().removeItem(itemStack);
                }
                if (clearedAmount >= itemCost) {
                    return true;
                }
            }
            return true;
        }
    }

    /**
     * Gets the total number of items contained in the given list of items
     *
     * @param items <p>The items to count</p>
     * @return <p>The total number of items</p>
     */
    private static int countItems(List<ItemStack> items) {
        int totalItems = 0;
        for (ItemStack itemStack : items) {
            totalItems += itemStack.getAmount();
        }
        return totalItems;
    }

    /**
     * Gets all empty books in a player's inventory
     *
     * @param player <p>The player to get books for</p>
     * @return <p>The empty books in the player's inventory</p>
     */
    private static List<ItemStack> getPlayersEmptyBooks(Player player) {
        List<ItemStack> validBooks = new ArrayList<>();
        for (ItemStack itemStack : player.getInventory().getContents()) {
            if (itemStack == null || itemStack.getType() != Material.WRITABLE_BOOK) {
                continue;
            }
            BookMeta book = (BookMeta) itemStack.getItemMeta();
            //Only accept empty books
            if (book != null && (!book.hasPages() || (book.getPageCount() == 1 && book.getPage(1).trim().isEmpty()))) {
                validBooks.add(itemStack);
            }
        }
        return validBooks;
    }

    /**
     * Uses economy to take payment for printing a number of books
     *
     * @param player    <p>The player which needs to pay</p>
     * @param cost      <p>The cost of the book printing</p>
     * @param numCopies <p>The number of books the player is printing</p>
     * @return <p>True if the player had the money and it has been withdrawn</p>
     */
    private static boolean payForBookPrintingEconomy(Player player, double cost, int numCopies) {
        if ((economy.getBalance(player) - cost) >= 0) {
            economy.withdrawPlayer(player, cost);
            BooksWithoutBorders.sendSuccessMessage(player, economy.format(cost) + " withdrawn to create " +
                    numCopies + " book(s)");
            BooksWithoutBorders.sendSuccessMessage(player, "New balance: " +
                    economy.format(economy.getBalance(player)));
            return true;
        } else {
            BooksWithoutBorders.sendErrorMessage(player, economy.format(cost) + " is required for this command!");
            return false;
        }
    }

    /**
     * Takes payment for printing a number of books by withdrawing the correct item
     *
     * @param player   <p>The player which needs to pay</p>
     * @param itemCost <p>The number of items to pay</p>
     */
    private static void payForBookPrintingItem(Player player, int itemCost) {
        PlayerInventory playerInventory = player.getInventory();

        int clearedAmount = 0;
        while (clearedAmount < itemCost) {
            int firstItemIndex = playerInventory.first(BooksWithoutBordersConfig.getBookPriceType());
            ItemStack firstItem = playerInventory.getItem(firstItemIndex);

            if (Objects.requireNonNull(firstItem).getAmount() <= itemCost - clearedAmount) {
                clearedAmount += firstItem.getAmount();
                player.getInventory().clear(firstItemIndex);
            } else {
                clearedAmount = itemCost;
                firstItem.setAmount(firstItem.getAmount() - (clearedAmount));
            }
        }
    }

}
